using System.Collections.Generic;
using System.Runtime.CompilerServices;
using MoonTools.ECS.Collections;

namespace MoonTools.ECS;

// TODO: we should implement a NativeDictionary that can be memcopied
public class Snapshot
{
	private Dictionary<TypeId, ComponentSnapshot> ComponentSnapshots = new Dictionary<TypeId, ComponentSnapshot>();

	private Dictionary<ArchetypeSignature, ArchetypeSnapshot> ArchetypeSnapshots =
		new Dictionary<ArchetypeSignature, ArchetypeSnapshot>();

	private Dictionary<TypeId, RelationSnapshot> RelationSnapshots =
		new Dictionary<TypeId, RelationSnapshot>();


	private Dictionary<Entity, ArchetypeRecord> EntityIndex = new Dictionary<Entity, ArchetypeRecord>();

	private Dictionary<Entity, IndexableSet<TypeId>> EntityRelationIndex =
		new Dictionary<Entity, IndexableSet<TypeId>>();

	private Dictionary<Entity, string> EntityTags = new Dictionary<Entity, string>();

	private IdAssigner EntityIdAssigner = new IdAssigner();

	public int Count
	{
		get
		{
			var count = 0;

			foreach (var snapshot in ArchetypeSnapshots.Values)
			{
				count += snapshot.Count;
			}

			return count;
		}
	}

	public void Restore(World world)
	{
		// restore archetype storage
		foreach (var (archetypeSignature, archetypeSnapshot) in ArchetypeSnapshots)
		{
			var archetype = world.ArchetypeIndex[archetypeSignature];
			archetypeSnapshot.Restore(archetype);
		}

		// restore entity index
		world.EntityIndex.Clear();
		foreach (var (id, ArchetypeRecord) in EntityIndex)
		{
			world.EntityIndex[id] = ArchetypeRecord;
		}

		// restore components
		foreach (var (typeId, componentSnapshot) in ComponentSnapshots)
		{
			var componentStorage = world.ComponentIndex[typeId];
			componentSnapshot.Restore(componentStorage);
		}

		// restore id assigner state
		EntityIdAssigner.CopyTo(world.EntityIdAssigner);

		// restore relation state
		foreach (var (typeId, relationSnapshot) in RelationSnapshots)
		{
			var relationStorage = world.RelationIndex[typeId];
			relationSnapshot.Restore(relationStorage);
		}

		// restore entity relation index state
		// FIXME: arghhhh this is so slow
		foreach (var (id, relationTypeSet) in EntityRelationIndex)
		{
			world.EntityRelationIndex[id].Clear();

			foreach (var typeId in relationTypeSet)
			{
				world.EntityRelationIndex[id].Add(typeId);
			}
		}

		foreach (var (id, s) in EntityTags)
		{
			world.EntityTags[id] = s;
		}
	}

	public void Take(World world)
	{
		// copy id assigner state
		world.EntityIdAssigner.CopyTo(EntityIdAssigner);

		// copy entity index
		EntityIndex.Clear();
		foreach (var (id, ArchetypeRecord) in world.EntityIndex)
		{
			EntityIndex[id] = ArchetypeRecord;
		}

		// copy archetypes
		foreach (var archetype in world.ArchetypeIndex.Values)
		{
			TakeArchetypeSnapshot(archetype);
		}

		// copy components
		foreach (var (typeId, componentStorage) in world.ComponentIndex)
		{
			TakeComponentSnapshot(typeId, componentStorage);
		}

		// copy relations
		foreach (var (typeId, relationStorage) in world.RelationIndex)
		{
			TakeRelationSnapshot(typeId, relationStorage);
		}

		// copy entity relation index
		// FIXME: arghhhh this is so slow
		foreach (var (id, relationTypeSet) in world.EntityRelationIndex)
		{
			if (!EntityRelationIndex.ContainsKey(id))
			{
				EntityRelationIndex.Add(id, new IndexableSet<TypeId>());
			}

			EntityRelationIndex[id].Clear();

			foreach (var typeId in relationTypeSet)
			{
				EntityRelationIndex[id].Add(typeId);
			}
		}

		foreach (var (id, s) in world.EntityTags)
		{
			EntityTags[id] = s;
		}
	}

	private void TakeArchetypeSnapshot(Archetype archetype)
	{
		if (!ArchetypeSnapshots.TryGetValue(archetype.Signature, out var archetypeSnapshot))
		{
			archetypeSnapshot = new ArchetypeSnapshot();
			ArchetypeSnapshots.Add(archetype.Signature, archetypeSnapshot);
		}

		archetypeSnapshot.Take(archetype);
	}

	private void TakeComponentSnapshot(TypeId typeId, ComponentStorage componentStorage)
	{
		if (!ComponentSnapshots.TryGetValue(typeId, out var componentSnapshot))
		{
			componentSnapshot = new ComponentSnapshot(componentStorage.Components.ElementSize);
			ComponentSnapshots.Add(typeId, componentSnapshot);
		}

		componentSnapshot.Take(componentStorage);
	}

	private void TakeRelationSnapshot(TypeId typeId, RelationStorage relationStorage)
	{
		if (!RelationSnapshots.TryGetValue(typeId, out var snapshot))
		{
			snapshot = new RelationSnapshot(relationStorage.relationDatas.ElementSize);
			RelationSnapshots.Add(typeId, snapshot);
		}

		snapshot.Take(relationStorage);
	}

	private class ArchetypeSnapshot
	{
		private readonly NativeArray<Entity> Entities;
		public int Count => Entities.Count;

		public ArchetypeSnapshot()
		{
			Entities = new NativeArray<Entity>();
		}

		public void Take(Archetype archetype)
		{
			archetype.Entities.CopyTo(Entities);
		}

		public void Restore(Archetype archetype)
		{
			Entities.CopyTo(archetype.Entities);
		}
	}

	private class ComponentSnapshot
	{
		private readonly Dictionary<Entity, int> EntityIDToStorageIndex = new Dictionary<Entity, int>();
		private readonly NativeArray Components;
		private readonly NativeArray<Entity> EntityIDs;

		public ComponentSnapshot(int elementSize)
		{
			Components = new NativeArray(elementSize);
			EntityIDs = new NativeArray<Entity>();
		}

		public void Take(ComponentStorage componentStorage)
		{
			componentStorage.Components.CopyAllTo(Components);
			componentStorage.EntityIDs.CopyTo(EntityIDs);
		}

		public void Restore(ComponentStorage componentStorage)
		{
			Components.CopyAllTo(componentStorage.Components);
			EntityIDs.CopyTo(componentStorage.EntityIDs);

			componentStorage.EntityIDToStorageIndex.Clear();
			for (int i = 0; i < EntityIDs.Count; i += 1)
			{
				var entityID = EntityIDs[i];
				componentStorage.EntityIDToStorageIndex[entityID] = i;
			}
		}
	}

	private class RelationSnapshot
	{
		private NativeArray Relations;
		private NativeArray RelationDatas;

		public RelationSnapshot(int elementSize)
		{
			Relations = new NativeArray(Unsafe.SizeOf<(Entity, Entity)>());
			RelationDatas = new NativeArray(elementSize);
		}

		public void Take(RelationStorage relationStorage)
		{
			relationStorage.relations.CopyAllTo(Relations);
			relationStorage.relationDatas.CopyAllTo(RelationDatas);
		}

		public void Restore(RelationStorage relationStorage)
		{
			relationStorage.Clear();

			Relations.CopyAllTo(relationStorage.relations);
			RelationDatas.CopyAllTo(relationStorage.relationDatas);

			for (int index = 0; index < Relations.Count; index += 1)
			{
				var relation = Relations.Get<(Entity, Entity)>(index);
				relationStorage.indices[relation] = index;

				relationStorage.indices[relation] = index;

				if (!relationStorage.outRelations.ContainsKey(relation.Item1))
				{
					relationStorage.outRelations[relation.Item1] =
						relationStorage.AcquireHashSetFromPool();
				}

				relationStorage.outRelations[relation.Item1].Add(relation.Item2);

				if (!relationStorage.inRelations.ContainsKey(relation.Item2))
				{
					relationStorage.inRelations[relation.Item2] =
						relationStorage.AcquireHashSetFromPool();
				}

				relationStorage.inRelations[relation.Item2].Add(relation.Item1);
			}
		}
	}
}
